local M = {}

local raceMarker = require("scenario/race_marker")
-- The pacenote core mod has a function to say stuff
-- TODO: Do the next thing with 'require'
local say = scripts_pacenotes_extension

local logTag = "rallyMode"
local isRally = true
local nextNote = {}
local last = 0
local prefix = ""
local suffix = ""

local hideMarkers = {}
local timeOffset = {}
local posOffset = {}
local cutoff = {}
local linkWords = {}
local breathLength = {}
local recce = {}
local allowedDists = {}
local slowCorners  = {}

-- From http://lua-users.org/wiki/CsvUtils
local function fromCSV(s)
  s = s .. ','        -- ending comma
  local t = {}        -- table to collect fields
  local fieldstart = 1
  repeat
    -- next field is quoted? (start with `"'?)
    if string.find(s, '^"', fieldstart) then
      local a, c
      local i  = fieldstart
      repeat
        -- find closing quote
        a, i, c = string.find(s, '"("?)', i+1)
      until c ~= '"'    -- quote not followed by quote?
      if not i then error('unmatched "') end
      local f = string.sub(s, fieldstart+1, i-1)
      table.insert(t, (string.gsub(f, '""', '"')))
      fieldstart = string.find(s, ',', i) + 1
    else                -- unquoted; find next comma
      local nexti = string.find(s, ',', fieldstart)
      table.insert(t, string.sub(s, fieldstart, nexti-1))
      fieldstart = nexti + 1
    end
  until fieldstart > string.len(s)
  return t
end

local function loadConfig()
  -- Reads config from ini file
  local c = loadIni("settings/rallyconfig.ini")
  hideMarkers = c.hideMarkers
  timeOffset = c.timeOffset
  posOffset = c.posOffset
  cutoff = c.cutoff
  linkWords = fromCSV(c.linkWords)
  breathLength = c.breathLength
  recce = c.recce
  allowedDists = fromCSV(c.allowedDists)
  slowCorners = fromCSV(c.slowCorners)
end

local function getWaypointName(index)
  local sc = scenario_scenarios.getScenario()
  return sc.BranchLapConfig[index]
  end

local function getMarker(n)
  -- Gets content of the "pacenote" DynamicField
  local name = getWaypointName(n)
  local obj = scenetree.findObject(name)
  local marker = obj:getDynDataFieldbyName("marker", 0)
  return marker
end

local function isSlow(s)
  for _, v in ipairs(slowCorners) do
    if s:find(v) then
      return true
    end
  end
end

local function isLinked(s)
  for _, v in ipairs(linkWords) do
    if v == s:match("(%w+)") then return true end
  end
end

local function distOrLink(d)
  -- Outputs rounded distance or linkWords when below cutoff
  local M = tonumber(allowedDists[#allowedDists])
  if d > M then
    return M
  end
  if d < cutoff then
    return ""
  else
    for i, v in ipairs(allowedDists) do
      -- The + 3 means we're rounding conservatively
      if d + 3 < tonumber(v) then
        return allowedDists[i-1]
      end
    end
  end
end

local function getWaypointPos(i)
  -- Gets the position vector of the i-th waypoint
  local name = getWaypointName(i)
  local obj = scenetree.findObject(name)
  return obj:getPosition()
end

local function getDistBtw(m, n)
  -- Returns dist between the m-th and n-th waypoints
  local d = 0
  for i=m, n-1, 1 do
    local a = getWaypointPos(i)
    local b = getWaypointPos(i+1)
    d = d + a:distance(b)
  end
  return d
end

local function getCall(i)
  -- Gets content of the "pacenote" DynamicField
  local name = getWaypointName(i)
  local obj = scenetree.findObject(name)
  local call = obj:getDynDataFieldbyName("pacenote", 0)
  return call
end

local function getPacenoteAfter(i)
  -- Gets next nonempty pacenote and its attributes
  local sc = scenario_scenarios.getScenario()
  local max = #sc.BranchLapConfig
  local inext = i + 1
  for k=inext, max, 1 do
    local pCall = getCall(k)
    if k >= max then
      -- TODO: Doesnt matter what's in the call. It won't say it.
      -- Maybe fix
      return { index = max, call = pCall }
    end
    if pCall ~= nil then
      if pCall ~= "empty" then
        return { index = k, call = pCall, slow = isSlow(pCall) }
      end
    end
  end
end

local function getPacenoteBefore(i)
  -- Gets the previous nonempty pacenote
  local icur = i
  for k=icur, 0, -1 do
    local pCall = getCall(k)
    if k == 0 then
      return { index = 0, call = pCall }
    end
    if pCall ~= nil then
      if pCall ~= "empty" then
        return { index = k, call = pCall }
      end
    end
  end
end

local function getDistCall(p)
  -- Gets the distance call to append to the next call
  local pFinal = getPacenoteAfter(p).index
  if (pFinal == p + 1) then
    local dist = getDistBtw(p, pFinal)
    return distOrLink(dist)
  elseif pFinal > p + 1 then
    for i=pFinal-1, p+1, -1 do
      if getMarker(i) ~= nil then
        p = i
      end
    end
    local dist = getDistBtw(p, pFinal)
    return distOrLink(dist)
   elseif pFinal <= p then
     log("E", logTag, "Next waypoint index is <= than the previous.")
   elseif pFinal == nil then
     log("E", logTag, "Could not get next waypoint index.")
  end
end

local function reset()
  loadConfig()
  if hideMarkers then raceMarker.hide(true) end
  last = 0
  local prefix = ""
  local suffix = ""
end

local function getWaypointId(index)
  -- TODO: This is probably stupidly roundabout
  local sc = scenario_scenarios.getScenario()
  local name = sc.BranchLapConfig[index]
  local obj = scenetree.findObject(name)
  return obj:getId()
end

local function getPlayerPos()
  return core_vehicles.getCurrentVehicleDetails().current.position
end

local function getPlayerVelocity()
  -- Returns the velocity vector. Components are in m/s
  local playerVehicle = scenetree.findObject(be:getPlayerVehicleID(0))
  return playerVehicle:getVelocity()
end

local function getPlayerSpeed()
  -- Returns the norm of the velocity vector
  local playerVehicle = scenetree.findObject(be:getPlayerVehicleID(0))
  local velocity = playerVehicle:getVelocity()
  return velocity:lengthGuarded()
end

local function getLastWaypointIndex()
  -- Returns the index of the last crossed waypoint
  local id = be:getPlayerVehicleID(0)
  local w = scenario_waypoints.state.vehicleWaypointsData[id]
  if w == nil then
    return 0
  end
  return w.cur
end

local function getDistFrom(n)
  -- Gets distance between player and n-th waypoint
  local i = getLastWaypointIndex() + 1
  local wPos = getWaypointPos(i)
  local pPos = getPlayerPos()
  local d = pPos:distance(wPos) + getDistBtw(i, n)
  return d
end

local function stageLength()
    local iFinal = #scenario_waypoints.state.originalBranches.mainPath
    return getDistBtw(1,iFinal)
end

local function getToFinish()
    local iFinal = #scenario_waypoints.state.originalBranches.mainPath
    return getDistFrom(iFinal)
end

local function getFromStart()
    return stageLength() - getToFinish()
end

local function getNextNextPacenote()
    local nextNote = getPacenoteAfter(last)
    local i = nextNote.index
    local nextNextNote = getPacenoteAfter(i)
    local nnIndex = nextNextNote.index
    local nnCall = getCall(nnIndex)
    local nnName = getWaypointName(nnIndex)
    return {index = nnIndex, call = nnCall, name = nnName }
end

local function getNextPacenote()
    local nextNote = getPacenoteAfter(last)
    local nIndex = nextNote.index
    local nCall = getCall(nIndex)
    local nName = getWaypointName(nIndex)
    return {index = nIndex, call = nCall, name = nName }
end

local function getPrevPacenote()
    local prevNote = getPacenoteBefore(last)
    local pIndex = prevNote.index
    local pCall = getCall(pIndex)
    local pName = getWaypointName(pIndex)
    return {index = pIndex, call = pCall, name = pName }
end

local function getLastWaypoint()
    local wIndex = getLastWaypointIndex()
    local wCall = getCall(wIndex)
    if wCall == nil then
        wCall = "empty"
    end
    local pName = getWaypointName(wIndex)
    return {index = wIndex, call = wCall, name = wName }
end

local function tokm(x)
    local xkm = x/1000
    return string.format("%.1f", xkm)
end

local function stats()
  -- Dumps useful stats for track-making
  local total = stageLength()
  local from = getFromStart()
  local to = getToFinish()
  local s = tostring(tokm(from) .. " / " .. tokm(total) .. " (km)")
  print(s)
  guihooks.trigger('Message', {ttl = 5, msg = s, category = "fill", icon = "flag"})
end

local function breathe(t)
  -- Takes a breath for n tenths
  for i=0, t, 1 do
    say.addPacenotes("p")
  end
end

-----------
-- Hooks --
-----------

local function onScenarioRestarted()
  if not isRally then return end
  reset()
end

local function onScenarioChange()
  if hideMarkers then raceMarker.hide(true) end
  if not isRally then return end
  loadConfig()
end

local function onRaceWaypointReached(data)
  if not isRally then return end
  if hideMarkers then raceMarker.hide(true) end
  local i = getLastWaypointIndex()
  if recce then
    local name = getWaypointName(i)
    print(name .. ' (' .. tostring(i) .. ')')
  end
end

local function onPreRender(dtReal, dtSim, dtRaw)
  if not isRally then return end
  local sc = scenario_scenarios.getScenario()
  if sc == nil then return end
  local raceState = sc.raceState
  local scenarioState = sc.state
  local max = #sc.BranchLapConfig
  if raceState == 'racing' and scenarioState == 'running' then
    local speed = getPlayerSpeed()
    local pred =  posOffset + speed * timeOffset
    local pnote = getPacenoteAfter(last)
    local i = pnote.index
    if i == nil then return end
    local dist = getDistFrom(i)
    if (dist < pred) and i > last and i < max then
      suffix = getDistCall(i)
      if recce then
        local name = getWaypointName(i)
        print(
          name
          .. ' ('
          .. tostring(i)
          .. '). CALL: '
          .. prefix
          .. ' '
          .. pnote.call
          .. ' '
          .. suffix
        )
      end
      say.addPacenotes(prefix .. ' ' .. pnote.call)
      if pnote.slow then
        breathe(round(timeOffset*10))
      else
        breathe(breathLength)
      end
      say.addPacenotes(tostring(suffix))
      breathe(breathLength)
      -- TODO: Implement the prefix override thing (see isLinked)
      if suffix == "" then
        prefix = linkWords[1]
      else
        prefix = ""
      end
      last = i
    end
  end
end

---------------------------
-- Testing and debugging --
---------------------------

local function info()
  dump(allowedDists)
  print("test")
end

local function utter(m,n)
  -- Utters m-th pacenote, followed by the next n pacenotes
  local prefix = ""
  for i=m, n, 1 do
    local pnote = getPacenoteAfter(i-1)
    local call = pnote.call
    local suffix = getDistCall(i)
    if recce then
      local name = getWaypointName(i)
      print(name .. ' (' .. tostring(i) .. '). CALL: ' .. prefix .. ' ' .. call .. ' ' .. suffix )
    end
    if suffix == "" then
      prefix = linkWords[1]
    else
      prefix = ""
    end
  end
end

local function utterPhrase(s)
  -- Says a phrase
  say.addPacenotes(s)
end

-- Loads a custom version of the scenario_waypoints extension
-- M.onRaceGoalsInitialised = function()
--   scenario_waypoints = extensions.scenario_waypointsNoSound
-- end

---------------
-- Interface --
---------------

M.onPreRender = onPreRender
M.onRaceWaypointReached = onRaceWaypointReached
M.onScenarioChange = onScenarioChange
M.onScenarioRestarted = onScenarioRestarted
M.onScenarioFinished = onScenarioFinished
M.onRaceTick = onRaceTick
M.isRally = isRally
M.stats = stats
-- Testing and debugging
M.info = info
M.utter = utter
M.utterPhrase = utterPhrase

return M
